"""Mask2Former representation"""
import json
from argparse import ArgumentParser, Namespace
from pathlib import Path
from datetime import datetime
from types import SimpleNamespace
from overrides import overrides
import torch as tr
import numpy as np
from lovely_tensors import monkey_patch
from fvcore.common.config import CfgNode

from vre.representations import Representation, ReprOut, LearnedRepresentationMixin
from vre.logger import vre_logger as logger
from vre.utils import (image_resize_batch, fetch_weights, image_read, image_write,
                       vre_load_weights, colorize_semantic_segmentation)

try:
    from .mask2former_impl import MaskFormer as MaskFormerImpl
except ImportError:
    from mask2former_impl import MaskFormer as MaskFormerImpl

monkey_patch()

def _get_output_shape(oldh: int, oldw: int, short_edge_length: int, max_size: int):
    """Compute the output size given input size and target short edge length."""
    scale = short_edge_length / min(oldh, oldw)
    newh, neww = (short_edge_length, scale * oldw) if oldh < oldw else (scale * oldh, short_edge_length)
    if max(newh, neww) > max_size:
        scale = max_size * 1.0 / max(newh, neww)
        newh, neww = newh * scale, neww * scale
    neww, newh = int(neww + 0.5), int(newh + 0.5)
    return newh, neww

class Mask2Former(Representation, LearnedRepresentationMixin):
    """Mask2Former representation implementation. Note: only semantic segmentation (not panoptic/instance) enabled."""
    def __init__(self, model_id: str, semantic_argmax_only: bool, **kwargs):
        super().__init__(**kwargs)
        assert isinstance(model_id, str) and model_id in {"47429163_0", "49189528_1", "49189528_0"}, model_id
        self._m2f_resources = Path(__file__).parent / "mask2former_impl/resources"
        self.classes, self.color_map, self.thing_dataset_id_to_contiguous_id = self._get_metadata(model_id)
        self.semantic_argmax_only = semantic_argmax_only
        self.model_id = model_id
        self.model: MaskFormerImpl | None = None
        self.cfg: CfgNode | None = None

    def vre_setup(self, load_weights = True):
        assert self.model is None, "vre_setup already called before"
        weights_path = fetch_weights(__file__) / f"{self.model_id}.ckpt"
        assert isinstance(weights_path, Path), type(weights_path)
        ckpt_data = vre_load_weights(weights_path)
        self.cfg = CfgNode(json.load(open(f"{self._m2f_resources}/{self.model_id}_cfg.json", "r")))
        params = MaskFormerImpl.from_config(self.cfg)
        params["metadata"] = SimpleNamespace(thing_dataset_id_to_contiguous_id=self.thing_dataset_id_to_contiguous_id)
        self.model = MaskFormerImpl(**{**params, "semantic_on": True, "panoptic_on": False, "instance_on": False})
        if load_weights:
            res = self.model.load_state_dict(ckpt_data["state_dict"], strict=False) # inference only: remove criterion
            assert res.unexpected_keys in (["criterion.empty_weight"], []), res
        self.model = self.model.eval().to(self.device)

    @tr.no_grad()
    @overrides
    def make(self, frames: np.ndarray, dep_data: dict[str, ReprOut] | None = None) -> ReprOut:
        height, width = frames.shape[1:3]
        _os = _get_output_shape(height, width, self.cfg.INPUT.MIN_SIZE_TEST, self.cfg.INPUT.MAX_SIZE_TEST)
        imgs = image_resize_batch(frames, _os[0], _os[1], "bilinear", "PIL").transpose(0, 3, 1, 2).astype("float32")
        inputs = [{"image": tr.from_numpy(img), "height": height, "width": width} for img in imgs]
        predictions: list[tr.Tensor] = [x["sem_seg"] for x in self.model(inputs)]
        res = []
        for pred in predictions:
            _pred = pred.argmax(0).byte() if self.semantic_argmax_only else pred.half().permute(1, 2, 0)
            res.append(_pred.to("cpu").numpy())
        return ReprOut(output=np.stack(res))

    @overrides
    def make_images(self, frames: np.ndarray, repr_data: ReprOut) -> np.ndarray:
        res = []
        frames_rsz = image_resize_batch(frames, *repr_data.output.shape[1:3])
        for img, pred in zip(frames_rsz, repr_data.output):
            _pred = pred if self.semantic_argmax_only else pred.argmax(-1)
            res.append(colorize_semantic_segmentation(_pred, self.classes, self.color_map, img))
        res = np.stack(res)
        return res

    @overrides
    def size(self, repr_data: ReprOut) -> tuple[int, int]:
        return repr_data.output.shape[1:3]

    @overrides
    def resize(self, repr_data: ReprOut, new_size: tuple[int, int]) -> ReprOut:
        interpolation = "nearest" if self.semantic_argmax_only else "bilinear"
        return ReprOut(output=image_resize_batch(repr_data.output, *new_size, interpolation=interpolation))

    def _get_metadata(self, model_id: str) -> tuple[list[str], list[tuple[int, int, int]], dict[str, int]]:
        metadata = None
        if model_id == "49189528_1":
            metadata = json.load(open(f"{self._m2f_resources}/mapillary_metadata.json", "r"))
        if model_id == "47429163_0":
            metadata = json.load(open(f"{self._m2f_resources}/coco_metadata.json", "r"))
        if model_id == "49189528_0":
            metadata = json.load(open(f"{self._m2f_resources}/mapillary_metadata2.json", "r"))
        return metadata["stuff_classes"], metadata["stuff_colors"], metadata.get("thing_dataset_id_to_contiguous_id")

    def vre_free(self):
        if str(self.device).startswith("cuda"):
            self.model.to("cpu")
            tr.cuda.empty_cache()
        self.model = None

def get_args() -> Namespace:
    """cli args"""
    parser = ArgumentParser()
    parser.add_argument("model_id", choices=["49189528_1", "47429163_0", "49189528_0"])
    parser.add_argument("input_image", type=Path)
    parser.add_argument("output_path", type=Path)
    return parser.parse_args()

def main(args: Namespace):
    """main fn. Usage: python mask2former.py 49189528_1/47429163_0/49189528_0 demo1.jpg output1.jpg"""
    img = image_read(args.input_image)

    m2f = Mask2Former(args.model_id, semantic_argmax_only=False, name="m2f", dependencies=[])
    m2f.device = "cuda" if tr.cuda.is_available() else "cpu"
    m2f.vre_setup()
    now = datetime.now()
    pred = m2f.make(img[None])
    logger.info(f"Pred took: {datetime.now() - now}")
    semantic_result: np.ndarray = m2f.make_images(img[None], pred)[0]
    image_write(semantic_result, args.output_path)
    logger.info(f"Written prediction to '{args.output_path}'")
    return semantic_result # for integration tests

if __name__ == "__main__":
    main(get_args())



# pylint: disable=all
import math
import numpy as np


def fov_diag_to_intrinsic(fov, res, res_new=None):
    """
    Ideal camera matrix for diagonal fov and image size.
    :param fov: Diagonal FOV in degrees
    :param res: (original_width, original_height) tuple
    :param res_new: (rescaled_width, rescaled_height) tuple, optional
    :return: Camera matrix
    """
    d = math.sqrt((res[0] / 2) ** 2 + (res[1] / 2) ** 2)
    z = d / math.tan(fov * math.pi / 180 / 2)
    if res_new is None:
        k = np.asarray([[z, 0, 0], [0, z, 0], [res[0] / 2, res[1] / 2, 1]])
    else:
        ax = math.atan(res[0] / 2 / z)
        ay = math.atan(res[1] / 2 / z)
        fx = res_new[0] / 2 / math.tan(ax)
        fy = res_new[1] / 2 / math.tan(ay)
        k = np.asarray([[fx, 0, 0], [0, fy, 0], [res_new[0] / 2, res_new[1] / 2, 1]])
    return k.transpose()


def fov_diag_crop_aspect(fov, ar, ar_new):
    """
    Camera fov on changing aspect ratio IFF the full image senzor is used (either in height or width).
    :param fov: Diagonal FOV in degrees
    :param ar: Original aspect ratio
    :param ar_new: New aspect ratio
    :return: New FOV
    """
    dd = diag(ar / ar_new, ar) / diag(ar, 1)
    return 2 * math.atan(dd * math.tan(fov / 2 * math.pi / 180)) * 180 / math.pi


def fov_hw_to_intrinsic(fov, res):
    """
    Ideal camera matrix for horizontal / vertical fov and image size.
    :param fov: (horizontal_fov, vertical_fov) tuple
    :param res: (height, width) tuple
    :return: Camera matrix
    """
    fx = res[0] / (2 * math.tan(1 / 2 * fov[0] * math.pi / 180))
    fy = res[1] / (2 * math.tan(1 / 2 * fov[1] * math.pi / 180))
    k = np.asarray([[fx, 0, 0], [0, fy, 0], [res[0]/2, res[1]/2, 1]])
    return k.transpose()


def fov_hw_to_focal(fov, res):
    """
    fx, fy for horizontal / vertical fov and image size.
    :param fov: (horizontal_fov, vertical_fov) tuple
    :param res: (height, width) tuple
    :return: fx, fy camera parameters
    """
    fx = res[0] / (2 * math.tan(1 / 2 * fov[0] * math.pi / 180))
    fy = res[1] / (2 * math.tan(1 / 2 * fov[1] * math.pi / 180))
    return fx, fy


def focal_from_35mm(f_equiv, crop_f):
    """
    Focal length from 35mm equivalent.
    :param f_equiv: 35mm equivalent length
    :param crop_f: Senzor crop factor
    :return:
    """
    return f_equiv / crop_f


def diag(x, y):
    return math.sqrt(x ** 2 + y ** 2)





"""NormedRepresentation module"""
import torch as tr

class NormedRepresentation:
    """NormedRepresentation implementation"""
    def __init__(self):
        self.normalization: str | None = None
        self.min: tr.Tensor | None = None
        self.max: tr.Tensor | None = None
        self.mean: tr.Tensor | None = None
        self.std: tr.Tensor | None = None

    def set_normalization(self, normalization: str, stats: tuple[tr.Tensor, tr.Tensor, tr.Tensor, tr.Tensor]):
        """sets the normalization"""
        assert normalization in ("min_max", "standardization"), normalization
        assert isinstance(stats, tuple) and len(stats) == 4, stats
        self.normalization = normalization
        self.min, self.max, self.mean, self.std = stats

    def normalize(self, x: tr.Tensor) -> tr.Tensor:
        """normalizes a data point read with self.load_from_disk(path) using external min/max information"""
        assert self.min is not None, "self.statistics must be set from reader before task.normalize(x)"
        return ((x.float() - self.min) / (self.max - self.min)).nan_to_num(0, 0, 0).float()

    def standardize(self, x: tr.Tensor) -> tr.Tensor:
        """standardizes a data point read with self.load_from_disk(path) using external min/max information"""
        assert self.min is not None, "self.statistics must be set from reader before task.normalize(x)"
        res = ((x.float() - self.mean) / self.std).nan_to_num(0, 0, 0)
        res[(res < -10) | (res > 10)] = 0
        return res.float()

    @property
    def stats(self) -> tuple[tr.Tensor, tr.Tensor, tr.Tensor, tr.Tensor] | None:
        """Returns a tuple with the min/max/mean/std tensors, or None if normalization is None"""
        if self.normalization is None:
            return None
        assert self.min is not None, f"Use self.set_normalization first for {self.normalization=}"
        return self.min, self.max, self.mean, self.std



"""Canny edge detector representation."""
import numpy as np
from overrides import overrides
from matplotlib.cm import gray # pylint: disable=no-name-in-module
from vre.representations import Representation, ReprOut
from vre.utils import image_resize_batch
from vre.utils.cv2_utils import cv2_Canny

class Canny(Representation):
    """Canny edge detector representation."""
    def __init__(self, threshold1: float, threshold2: float, aperture_size: int, l2_gradient: bool, **kwargs):
        super().__init__(**kwargs)
        self.threshold1 = threshold1
        self.threshold2 = threshold2
        self.aperture_size = aperture_size
        self.l2_gradient = l2_gradient

    @overrides
    def make(self, frames: np.ndarray, dep_data: dict[str, ReprOut] | None = None) -> ReprOut:
        return ReprOut(output=np.array([self._make_one(frame) for frame in frames]))

    @overrides
    def make_images(self, frames: np.ndarray, repr_data: ReprOut) -> np.ndarray:
        return (255 * gray(repr_data.output)[..., 0:3]).astype(np.uint8)

    @overrides
    def size(self, repr_data: ReprOut) -> tuple[int, int]:
        return repr_data.output.shape[1:3]

    @overrides
    def resize(self, repr_data: ReprOut, new_size: tuple[int, int]) -> ReprOut:
        return ReprOut(output=image_resize_batch(repr_data.output, *new_size))

    def _make_one(self, x: np.ndarray) -> np.ndarray:
        res = cv2_Canny(x, threshold1=self.threshold1, threshold2=self.threshold2,
                        apertureSize=self.aperture_size, L2gradient=self.l2_gradient)
        res = np.float32(res) / 255
        return res



"""Depth normals representation using SVD."""
from overrides import overrides
import numpy as np

from vre.representations.normals.depth_svd.depth_svd_impl import \
    fov_diag_to_intrinsic, get_sampling_grid, get_normalized_coords, depth_to_normals
from vre.representations import Representation, ReprOut
from vre.utils import image_resize_batch

class DepthNormalsSVD(Representation):
    """
    General method for estimating normals from a depth map (+ intrinsics): a 2D window centered on each pixel is
    projected into 3D and then a plane is fitted on the 3D pointcloud using SVD.
    """
    def __init__(self, sensor_fov: int, sensor_width: int, sensor_height: int, window_size: int,
                 input_downsample_step: int = None, stride: int = None, max_distance: float = None,
                 min_valid_count: int = None, **kwargs):
        assert window_size % 2 == 1, f"Expected odd window size. Got: {window_size}"
        self.sensor_fov = sensor_fov
        self.sensor_width = sensor_width
        self.sensor_height = sensor_height
        self.window_size = window_size
        self.stride = stride if stride is not None else 1
        self.input_downsample_step = input_downsample_step if input_downsample_step is not None else 1
        self.max_dist = max_distance if max_distance is not None else -1
        self.min_valid = min_valid_count if min_valid_count is not None else 0
        super().__init__(**kwargs)
        assert len(self.dependencies) == 1, "Expected one depth method!"
        self._grid_cache = {}

    @overrides
    def make(self, frames: np.ndarray, dep_data: dict[str, ReprOut] | None = None) -> ReprOut:
        depths = dep_data[self.dependencies[0].name].output
        assert len(depths.shape) == 3, f"Expected (T, H, W) got: {depths.shape}"
        res = np.array([self._make_one_normal(depth) for depth in depths])
        return ReprOut(output=res)

    @overrides
    def make_images(self, frames: np.ndarray, repr_data: ReprOut) -> np.ndarray:
        return (repr_data.output * 255).astype(np.uint8)

    @overrides
    def size(self, repr_data: ReprOut) -> tuple[int, int]:
        return repr_data.output.shape[1:3]

    @overrides
    def resize(self, repr_data: ReprOut, new_size: tuple[int, int]) -> ReprOut:
        return ReprOut(output=image_resize_batch(repr_data.output, *new_size))

    def _make_one_normal(self, depth: np.ndarray) -> np.ndarray:
        # TODO: batch vectorize this if possible
        sampling_grid, normalized_grid = self._get_grid(depth)
        if self.input_downsample_step is not None:
            depth = depth[:: self.input_downsample_step, :: self.input_downsample_step]
        normals = depth_to_normals(depth, sampling_grid, normalized_grid, self.max_dist, self.min_valid)
        normals = (normals.astype(np.float32) + 1) / 2
        return normals

    def _get_grid(self, depth: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        height, width = depth.shape[:2]
        if (height, width) in self._grid_cache:
            return self._grid_cache[(height, width)]
        if self.input_downsample_step is not None:
            depth = depth[:: self.input_downsample_step, :: self.input_downsample_step]
        depth_height, depth_width = depth.shape[:2]
        sampling_grid = get_sampling_grid(depth_width, depth_height, self.window_size, self.stride)
        K = fov_diag_to_intrinsic(self.sensor_fov, (self.sensor_width, self.sensor_height), (depth_width, depth_height))
        normalized_grid = get_normalized_coords(depth_width, depth_height, K)
        self._grid_cache[(height, width)] = sampling_grid, normalized_grid
        return sampling_grid, normalized_grid



from __future__ import annotations
from pathlib import Path
import numpy as np
import torch as tr
import flow_vis
from overrides import overrides
from matplotlib.cm import Spectral # pylint: disable=no-name-in-module
from torch.nn import functional as F

from ..representations.hsv import rgb2hsv
from ..utils import colorize_semantic_segmentation
from ..logger import vre_logger as logger
from .stored_represntation import NpzRepresentation
from .normed_representation import NormedRepresentation

class ColorRepresentation(NpzRepresentation, NormedRepresentation):
    @overrides
    def plot_fn(self, x: tr.Tensor) -> np.ndarray:
        """to be removed"""
        logger.warning("DELETE THIS")
        min, max, mean, std = self.stats # pylint: disable=all
        assert isinstance(x, tr.Tensor), type(x)
        assert len(x.shape) == 3, x.shape # guaranteed to be (H, W, C) at this point
        x = x.nan_to_num(0).cpu().detach()
        if x.shape[-1] != 3:
            x = x[..., 0:1]
        if x.shape[-1] == 1: # guaranteed to be (H, W, 3) after this if statement hopefully
            x = x.repeat(1, 1, 3)
        if x.dtype == tr.uint8 or self.is_classification:
            return x.numpy()
        if self.normalization is not None:
            x = (x * std + mean) if self.normalization == "standardization" else x
            x = x * (max - min) + min if self.normalization == "min_max" else x
            x = (x * 255) if (max <= 1).any() else x
        x = x.numpy().astype(np.uint8)
        return x

class RGBRepresentation(ColorRepresentation):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, n_channels=3, **kwargs)

class HSVRepresentation(RGBRepresentation):
    @overrides
    def load_from_disk(self, path: Path) -> tr.Tensor:
        rgb = super().load_from_disk(path)
        return tr.from_numpy(rgb2hsv(rgb.numpy())).float()

class EdgesRepresentation(ColorRepresentation):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, n_channels=1, **kwargs)

class NormalsRepresentation(ColorRepresentation):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, n_channels=3, **kwargs)

class DepthRepresentation(NpzRepresentation, NormedRepresentation):
    """DepthRepresentation. Implements depth task-specific stuff, like spectral map for plots."""
    def __init__(self, name: str, min_depth: float, max_depth: float, *args, **kwargs):
        super().__init__(name, n_channels=1, *args, **kwargs)
        self.min_depth = min_depth
        self.max_depth = max_depth

    @overrides
    def load_from_disk(self, path: Path) -> tr.Tensor:
        """Reads the npz data from the disk and transforms it properly"""
        res = super().load_from_disk(path)
        res_clip = res.clip(self.min_depth, self.max_depth)
        return res_clip

    @overrides
    def plot_fn(self, x: tr.Tensor) -> np.ndarray:
        x = x.detach().clip(0, 1).squeeze().cpu().numpy()
        _min, _max = np.percentile(x, [1, 95])
        x = np.nan_to_num((x - _min) / (_max - _min), False, 0, 0, 0).clip(0, 1)
        y: np.ndarray = Spectral(x)[..., 0:3] * 255
        return y.astype(np.uint8)

class OpticalFlowRepresentation(NpzRepresentation, NormedRepresentation):
    """OpticalFlowRepresentation. Implements flow task-specific stuff, like using flow_vis."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, n_channels=2, **kwargs)

    @overrides
    def plot_fn(self, x: tr.Tensor) -> np.ndarray:
        _min, _max = x.min(0)[0].min(0)[0], x.max(0)[0].max(0)[0]
        x = ((x - _min) / (_max - _min)).nan_to_num(0, 0, 0).detach().cpu().numpy()
        return flow_vis.flow_to_color(x)

class SemanticRepresentation(NpzRepresentation, NormedRepresentation): # TODO: no norm
    """SemanticRepresentation. Implements semantic task-specific stuff, like argmaxing if needed"""
    def __init__(self, *args, classes: int | list[str], color_map: list[tuple[int, int, int]], **kwargs):
        self.n_classes = len(list(range(classes)) if isinstance(classes, int) else classes)
        super().__init__(*args, **kwargs, n_channels=self.n_classes)
        self.classes = list(range(classes)) if isinstance(classes, int) else classes
        self.color_map = color_map
        assert len(color_map) == self.n_classes and self.n_classes > 1, (color_map, self.n_classes)

    @overrides
    def load_from_disk(self, path: Path) -> tr.Tensor:
        res = super().load_from_disk(path)
        if len(res.shape) == 3:
            assert res.shape[-1] == self.n_classes, f"Expected {self.n_classes} (HxWxC), got {res.shape[-1]}"
            res = res.argmax(-1)
        assert len(res.shape) == 2, f"Only argmaxed data supported, got: {res.shape}"
        res = F.one_hot(res.long(), num_classes=self.n_classes).float()
        return res

    @overrides
    def plot_fn(self, x: tr.Tensor) -> np.ndarray:
        x_argmax = x.squeeze().nan_to_num(0).detach().argmax(-1).cpu().numpy()
        return colorize_semantic_segmentation(x_argmax, self.classes, self.color_map,
                                              font_size_scale=2, original_rgb=None)



"""SafeUAV semanetic segmentation representation"""
from overrides import overrides
import numpy as np
import torch as tr
from torch import nn
from torch.nn import functional as F
from vre.representations import Representation, ReprOut, LearnedRepresentationMixin
from vre.utils import image_resize_batch, fetch_weights, vre_load_weights, colorize_semantic_segmentation
from vre.logger import vre_logger as logger

try:
    from .Map2Map import EncoderMap2Map, DecoderMap2Map
except ImportError:
    from Map2Map import EncoderMap2Map, DecoderMap2Map

class _SafeUavWrapper(nn.Module):
    """Wrapper. TODO: Replace with nn.Sequential"""
    def __init__(self, ch_in: int, ch_out: int):
        super().__init__()
        tr.manual_seed(42)
        self.encoder = EncoderMap2Map(ch_in)
        self.decoder = DecoderMap2Map(ch_out)

    def forward(self, x):
        """forward function"""
        y_encoder = self.encoder(x)
        y_decoder = self.decoder(y_encoder)
        return y_decoder

# TODO: make semantic_argmax_only not optional
class SafeUAV(Representation, LearnedRepresentationMixin):
    """SafeUAV semantic segmentation representation"""
    def __init__(self, num_classes: int, train_height: int, train_width: int,
                 color_map: list[tuple[int, int, int]], semantic_argmax_only: bool = True,
                 weights_file: str | None = None, **kwargs):
        super().__init__(**kwargs)
        self.num_classes = num_classes
        assert len(color_map) == num_classes, f"{color_map} ({len(color_map)}) vs {num_classes}"
        self.color_map = color_map
        self.train_height = train_height
        self.train_width = train_width
        self.semantic_argmax_only = semantic_argmax_only
        self.weights_file = weights_file
        self.classes = list(range(num_classes))
        self.model: _SafeUavWrapper | None = None

    def vre_setup(self, load_weights: bool = True):
        self.model = _SafeUavWrapper(ch_in=3, ch_out=self.num_classes)

        if load_weights:
            if self.weights_file is None:
                logger.warning("No weights file provided, using random weights.")
                return

            def _convert(data: dict[str, tr.Tensor]) -> dict[str, tr.Tensor]:
                logger.warning("GET RID OF THIS WHEN THERE'S TIME")
                new_data = {}
                for k in data.keys():
                    if k.startswith("model.0."):
                        other = k.replace("model.0.", "encoder.")
                    elif k.startswith("model.1."):
                        other = k.replace("model.1.", "decoder.")
                    else:
                        assert False, k
                    new_data[other] = data[k]
                return new_data

            weights_file_abs = fetch_weights(__file__) / self.weights_file
            data = _convert(vre_load_weights(weights_file_abs)["state_dict"])
            self.model.load_state_dict(data)
        self.model = self.model.eval().to(self.device)

    @overrides
    def make(self, frames: np.ndarray, dep_data: dict[str, ReprOut] | None = None) -> ReprOut:
        tr_frames = tr.from_numpy(frames).to(self.device)
        frames_norm = tr_frames.permute(0, 3, 1, 2) / 255
        frames_resized = F.interpolate(frames_norm, (self.train_height, self.train_width), mode="bilinear")
        with tr.no_grad():
            prediction = self.model.forward(frames_resized)
        np_pred = prediction.permute(0, 2, 3, 1).cpu().numpy()
        y_out = np.argmax(np_pred, axis=-1).astype(np.uint8) if self.semantic_argmax_only else np_pred
        return ReprOut(output=y_out)

    @overrides
    def make_images(self, frames: np.ndarray, repr_data: ReprOut) -> np.ndarray:
        res = []
        frames_rsz = image_resize_batch(frames, *repr_data.output.shape[1:3])
        for img, pred in zip(frames_rsz, repr_data.output):
            _pred = pred if self.semantic_argmax_only else pred.argmax(-1)
            res.append(colorize_semantic_segmentation(_pred, self.classes, self.color_map, img))
        res = np.stack(res)
        return res

    @overrides
    def size(self, repr_data: ReprOut) -> tuple[int, int]:
        return repr_data.output.shape[1:3]

    @overrides
    def resize(self, repr_data: ReprOut, new_size: tuple[int, int]) -> ReprOut:
        interpolation = "nearest" if self.semantic_argmax_only else "bilinear"
        return ReprOut(image_resize_batch(repr_data.output, *new_size, interpolation=interpolation))

    def vre_free(self):
        if str(self.device).startswith("cuda"):
            self.model.to("cpu")
            tr.cuda.empty_cache()



"""ensemble module for marigold"""
from functools import partial

import numpy as np
import torch
from torchvision.transforms import InterpolationMode
from vre.logger import vre_logger as logger

from .marigold_util import resize_max_res

try:
    import scipy
except ImportError:
    logger.warning("Scipy is not installed, cannot use ensembles")

# pylint: disable=too-many-statements
def ensemble_depth(
    depth: torch.Tensor,
    scale_invariant: bool = True,
    shift_invariant: bool = True,
    output_uncertainty: bool = False,
    reduction: str = "median",
    regularizer_strength: float = 0.02,
    max_iter: int = 2,
    tol: float = 1e-3,
    max_res: int = 1024,
) -> tuple[torch.Tensor, torch.Tensor | None]:
    """
    Ensembles depth maps represented by the `depth` tensor with expected shape `(B, 1, H, W)`, where B is the
    number of ensemble members for a given prediction of size `(H x W)`. Even though the function is designed for
    depth maps, it can also be used with disparity maps as long as the input tensor values are non-negative. The
    alignment happens when the predictions have one or more degrees of freedom, that is when they are either
    affine-invariant (`scale_invariant=True` and `shift_invariant=True`), or just scale-invariant (only
    `scale_invariant=True`). For absolute predictions (`scale_invariant=False` and `shift_invariant=False`)
    alignment is skipped and only ensembling is performed.

    Args:
        depth (`torch.Tensor`):
            Input ensemble depth maps.
        scale_invariant (`bool`, *optional*, defaults to `True`):
            Whether to treat predictions as scale-invariant.
        shift_invariant (`bool`, *optional*, defaults to `True`):
            Whether to treat predictions as shift-invariant.
        output_uncertainty (`bool`, *optional*, defaults to `False`):
            Whether to output uncertainty map.
        reduction (`str`, *optional*, defaults to `"median"`):
            Reduction method used to ensemble aligned predictions. The accepted values are: `"mean"` and
            `"median"`.
        regularizer_strength (`float`, *optional*, defaults to `0.02`):
            Strength of the regularizer that pulls the aligned predictions to the unit range from 0 to 1.
        max_iter (`int`, *optional*, defaults to `2`):
            Maximum number of the alignment solver steps. Refer to `scipy.optimize.minimize` function, `options`
            argument.
        tol (`float`, *optional*, defaults to `1e-3`):
            Alignment solver tolerance. The solver stops when the tolerance is reached.
        max_res (`int`, *optional*, defaults to `1024`):
            Resolution at which the alignment is performed; `None` matches the `processing_resolution`.
    Returns:
        A tensor of aligned and ensembled depth maps and optionally a tensor of uncertainties of the same shape:
        `(1, 1, H, W)`.
    """
    if depth.dim() != 4 or depth.shape[1] != 1:
        raise ValueError(f"Expecting 4D tensor of shape [B,1,H,W]; got {depth.shape}.")
    if reduction not in ("mean", "median"):
        raise ValueError(f"Unrecognized reduction method: {reduction}.")
    if not scale_invariant and shift_invariant:
        raise ValueError("Pure shift-invariant ensembling is not supported.")

    def init_param(depth: torch.Tensor):
        init_min = depth.reshape(ensemble_size, -1).min(dim=1).values
        init_max = depth.reshape(ensemble_size, -1).max(dim=1).values

        if scale_invariant and shift_invariant:
            init_s = 1.0 / (init_max - init_min).clamp(min=1e-6)
            init_t = -init_s * init_min
            param = torch.cat((init_s, init_t)).cpu().numpy()
        elif scale_invariant:
            init_s = 1.0 / init_max.clamp(min=1e-6)
            param = init_s.cpu().numpy()
        else:
            raise ValueError("Unrecognized alignment.")

        return param

    def align(depth: torch.Tensor, param: np.ndarray) -> torch.Tensor:
        if scale_invariant and shift_invariant:
            s, t = np.split(param, 2)
            s = torch.from_numpy(s).to(depth).view(ensemble_size, 1, 1, 1)
            t = torch.from_numpy(t).to(depth).view(ensemble_size, 1, 1, 1)
            out = depth * s + t
        elif scale_invariant:
            s = torch.from_numpy(param).to(depth).view(ensemble_size, 1, 1, 1)
            out = depth * s
        else:
            raise ValueError("Unrecognized alignment.")
        return out

    def ensemble(depth_aligned: torch.Tensor,
                 return_uncertainty: bool = False) -> tuple[torch.Tensor, torch.Tensor | None]:
        uncertainty = None
        if reduction == "mean":
            prediction = torch.mean(depth_aligned, dim=0, keepdim=True)
            if return_uncertainty:
                uncertainty = torch.std(depth_aligned, dim=0, keepdim=True)
        elif reduction == "median":
            prediction = torch.median(depth_aligned, dim=0, keepdim=True).values
            if return_uncertainty:
                uncertainty = torch.median(torch.abs(depth_aligned - prediction), dim=0, keepdim=True).values
        else:
            raise ValueError(f"Unrecognized reduction method: {reduction}.")
        return prediction, uncertainty

    def cost_fn(param: np.ndarray, depth: torch.Tensor) -> float:
        cost = 0.0
        depth_aligned = align(depth, param)

        for i, j in torch.combinations(torch.arange(ensemble_size)):
            diff = depth_aligned[i] - depth_aligned[j]
            cost += (diff**2).mean().sqrt().item()

        if regularizer_strength > 0:
            prediction, _ = ensemble(depth_aligned, return_uncertainty=False)
            err_near = (0.0 - prediction.min()).abs().item()
            err_far = (1.0 - prediction.max()).abs().item()
            cost += (err_near + err_far) * regularizer_strength

        return cost

    def compute_param(depth: torch.Tensor):
        depth_to_align = depth.to(torch.float32)
        if max_res is not None and max(depth_to_align.shape[2:]) > max_res:
            depth_to_align = resize_max_res(depth_to_align, max_res, InterpolationMode.NEAREST_EXACT)

        param = init_param(depth_to_align)
        res = scipy.optimize.minimize(partial(cost_fn, depth=depth_to_align), param, method="BFGS",
                                      tol=tol, options={"maxiter": max_iter, "disp": False})
        return res.x

    requires_aligning = scale_invariant or shift_invariant
    ensemble_size = depth.shape[0]

    if requires_aligning:
        param = compute_param(depth)
        depth = align(depth, param)

    depth, uncertainty = ensemble(depth, return_uncertainty=output_uncertainty)

    depth_max = depth.max()
    if scale_invariant and shift_invariant:
        depth_min = depth.min()
    elif scale_invariant:
        depth_min = 0
    else:
        raise ValueError("Unrecognized alignment.")
    depth_range = (depth_max - depth_min).clamp(min=1e-6)
    depth = (depth - depth_min) / depth_range
    if output_uncertainty:
        uncertainty /= depth_range

    return depth, uncertainty  # [1,1,H,W], [1,1,H,W]



"""init file"""
from .utils import (get_project_root, parsed_str_type, took, is_dir_empty, get_closest_square,
                    now_fmt, is_git_lfs, semantic_mapper)
from .vre_video import FakeVideo, VREVideo
from .topological_sort import topological_sort
from .image import (image_resize, image_resize_batch, image_write, image_read, image_add_title, collage_fn,
                    image_add_border)
from .resources import fetch_weights, vre_load_weights, fetch_resource
from .batches import make_batches, all_batch_exists
from .colorizer import colorize_depth, colorize_semantic_segmentation, colorize_optical_flow



# pylint: disable=all
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from model.warplayer import warp

def conv_wo_act(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=False),
        nn.BatchNorm2d(out_planes),
    )


def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=False),
        nn.BatchNorm2d(out_planes),
        nn.PReLU(out_planes)
    )


class ResBlock(nn.Module):
    def __init__(self, in_planes, out_planes, stride=1):
        super(ResBlock, self).__init__()
        if in_planes == out_planes and stride == 1:
            self.conv0 = nn.Identity()
        else:
            self.conv0 = nn.Conv2d(in_planes, out_planes,
                                   3, stride, 1, bias=False)
        self.conv1 = conv(in_planes, out_planes, 5, stride, 2)
        self.conv2 = conv_wo_act(out_planes, out_planes, 3, 1, 1)
        self.relu1 = nn.PReLU(1)
        self.relu2 = nn.PReLU(out_planes)
        self.fc1 = nn.Conv2d(out_planes, 16, kernel_size=1, bias=False)
        self.fc2 = nn.Conv2d(16, out_planes, kernel_size=1, bias=False)

    def forward(self, x):
        y = self.conv0(x)
        x = self.conv1(x)
        x = self.conv2(x)
        w = x.mean(3, True).mean(2, True)
        w = self.relu1(self.fc1(w))
        w = torch.sigmoid(self.fc2(w))
        x = self.relu2(x * w + y)
        return x


class IFBlock(nn.Module):
    def __init__(self, in_planes, scale=1, c=64):
        super(IFBlock, self).__init__()
        self.scale = scale
        self.conv0 = conv(in_planes, c, 5, 2, 2)
        self.res0 = ResBlock(c, c)
        self.res1 = ResBlock(c, c)
        self.res2 = ResBlock(c, c)
        self.res3 = ResBlock(c, c)
        self.res4 = ResBlock(c, c)
        self.res5 = ResBlock(c, c)
        self.conv1 = nn.Conv2d(c, 8, 3, 1, 1)
        self.up = nn.PixelShuffle(2)

    def forward(self, x):
        if self.scale != 1:
            x = F.interpolate(x, scale_factor=1. / self.scale, mode="bilinear",
                              align_corners=False)
        x = self.conv0(x)
        x = self.res0(x)
        x = self.res1(x)
        x = self.res2(x)
        x = self.res3(x)
        x = self.res4(x)
        x = self.res5(x)
        x = self.conv1(x)
        flow = self.up(x)
        if self.scale != 1:
            flow = F.interpolate(flow, scale_factor=self.scale, mode="bilinear",
                                 align_corners=False)
        return flow


class IFNet(nn.Module):
    def __init__(self):
        super(IFNet, self).__init__()
        self.block0 = IFBlock(6, scale=8, c=192)
        self.block1 = IFBlock(8, scale=4, c=128)
        self.block2 = IFBlock(8, scale=2, c=96)
        self.block3 = IFBlock(8, scale=1, c=48)

    def forward(self, x, UHD=False):
        if UHD:
            x = F.interpolate(x, scale_factor=0.25, mode="bilinear", align_corners=False)
        else:
            x = F.interpolate(x, scale_factor=0.5, mode="bilinear",
                              align_corners=False)
        flow0 = self.block0(x)
        F1 = flow0
        warped_img0 = warp(x[:, :3], F1)
        warped_img1 = warp(x[:, 3:], -F1)
        flow1 = self.block1(torch.cat((warped_img0, warped_img1, F1), 1))
        F2 = (flow0 + flow1)
        warped_img0 = warp(x[:, :3], F2)
        warped_img1 = warp(x[:, 3:], -F2)
        flow2 = self.block2(torch.cat((warped_img0, warped_img1, F2), 1))
        F3 = (flow0 + flow1 + flow2)
        warped_img0 = warp(x[:, :3], F3)
        warped_img1 = warp(x[:, 3:], -F3)
        flow3 = self.block3(torch.cat((warped_img0, warped_img1, F3), 1))
        F4 = (flow0 + flow1 + flow2 + flow3)
        return F4, [F1, F2, F3, F4]

if __name__ == "__main__":
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    img0 = torch.zeros(3, 3, 256, 256).float().to(device)
    img1 = torch.tensor(np.random.normal(
        0, 1, (3, 3, 256, 256))).float().to(device)
    imgs = torch.cat((img0, img1), 1)
    flownet = IFNet()
    flow, _ = flownet(imgs)
    print(flow.shape)



# pylint: disable=all
import time
import torch
import numpy as np
import torchvision
from torch.nn import functional as F
from vre.logger import vre_logger as logger

def scale_boxes(img1_shape, boxes, img0_shape, ratio_pad=None):
    """
    Rescales bounding boxes (in the format of xyxy) from the shape of the image they were originally specified in
    (img1_shape) to the shape of a different image (img0_shape).

    Args:
      img1_shape (tuple): The shape of the image that the bounding boxes are for, in the format of (height, width).
      boxes (torch.Tensor): the bounding boxes of the objects in the image, in the format of (x1, y1, x2, y2)
      img0_shape (tuple): the shape of the target image, in the format of (height, width).
      ratio_pad (tuple): a tuple of (ratio, pad) for scaling the boxes. If not provided, the ratio and pad will be
                         calculated based on the size difference between the two images.

    Returns:
      boxes (torch.Tensor): The scaled bounding boxes, in the format of (x1, y1, x2, y2)
    """
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
        pad = round((img1_shape[1] - img0_shape[1] * gain) / 2 - 0.1), round(
            (img1_shape[0] - img0_shape[0] * gain) / 2 - 0.1)  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    boxes[..., [0, 2]] -= pad[0]  # x padding
    boxes[..., [1, 3]] -= pad[1]  # y padding
    boxes[..., :4] /= gain
    clip_boxes(boxes, img0_shape)
    return boxes

def clip_boxes(boxes, shape):
    """
    It takes a list of bounding boxes and a shape (height, width) and clips the bounding boxes to the
    shape

    Args:
      boxes (torch.Tensor): the bounding boxes to clip
      shape (tuple): the shape of the image
    """
    if isinstance(boxes, torch.Tensor):  # faster individually
        boxes[..., 0].clamp_(0, shape[1])  # x1
        boxes[..., 1].clamp_(0, shape[0])  # y1
        boxes[..., 2].clamp_(0, shape[1])  # x2
        boxes[..., 3].clamp_(0, shape[0])  # y2
    else:  # np.array (faster grouped)
        boxes[..., [0, 2]] = boxes[..., [0, 2]].clip(0, shape[1])  # x1, x2
        boxes[..., [1, 3]] = boxes[..., [1, 3]].clip(0, shape[0])  # y1, y2


def non_max_suppression(
        prediction,
        conf_thres=0.25,
        iou_thres=0.45,
        classes=None,
        agnostic=False,
        multi_label=False,
        labels=(),
        max_det=300,
        nc=0,  # number of classes (optional)
        max_time_img=0.05,
        max_nms=30000,
        max_wh=7680,
):
    """
    Perform non-maximum suppression (NMS) on a set of boxes, with support for masks and multiple labels per box.

    Arguments:
        prediction (torch.Tensor): A tensor of shape (batch_size, num_classes + 4 + num_masks, num_boxes)
            containing the predicted boxes, classes, and masks. The tensor should be in the format
            output by a model, such as YOLO.
        conf_thres (float): The confidence threshold below which boxes will be filtered out.
            Valid values are between 0.0 and 1.0.
        iou_thres (float): The IoU threshold below which boxes will be filtered out during NMS.
            Valid values are between 0.0 and 1.0.
        classes (List[int]): A list of class indices to consider. If None, all classes will be considered.
        agnostic (bool): If True, the model is agnostic to the number of classes, and all
            classes will be considered as one.
        multi_label (bool): If True, each box may have multiple labels.
        labels (List[List[Union[int, float, torch.Tensor]]]): A list of lists, where each inner
            list contains the apriori labels for a given image. The list should be in the format
            output by a dataloader, with each label being a tuple of (class_index, x1, y1, x2, y2).
        max_det (int): The maximum number of boxes to keep after NMS.
        nc (int, optional): The number of classes output by the model. Any indices after this will be considered masks.
        max_time_img (float): The maximum time (seconds) for processing one image.
        max_nms (int): The maximum number of boxes into torchvision.ops.nms().
        max_wh (int): The maximum box width and height in pixels

    Returns:
        (List[torch.Tensor]): A list of length batch_size, where each element is a tensor of
            shape (num_boxes, 6 + num_masks) containing the kept boxes, with columns
            (x1, y1, x2, y2, confidence, class, mask1, mask2, ...).
    """

    # Checks
    assert 0 <= conf_thres <= 1, f'Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0'
    assert 0 <= iou_thres <= 1, f'Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0'
    if isinstance(prediction, (list, tuple)):  # YOLOv8 model in validation model, output = (inference_out, loss_out)
        prediction = prediction[0]  # select only inference output

    device = prediction.device
    mps = 'mps' in device.type  # Apple MPS
    if mps:  # MPS not fully supported yet, convert tensors to CPU before NMS
        prediction = prediction.cpu()
    bs = prediction.shape[0]  # batch size
    nc = nc or (prediction.shape[1] - 4)  # number of classes
    nm = prediction.shape[1] - nc - 4
    mi = 4 + nc  # mask start index
    xc = prediction[:, 4:mi].amax(1) > conf_thres  # candidates

    # Settings
    # min_wh = 2  # (pixels) minimum box width and height
    time_limit = 0.5 + max_time_img * bs  # seconds to quit after
    redundant = True  # require redundant detections
    multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)
    merge = False  # use merge-NMS

    t = time.time()
    output = [torch.zeros((0, 6 + nm), device=prediction.device)] * bs
    for xi, x in enumerate(prediction):  # image index, image inference
        # Apply constraints
        # x[((x[:, 2:4] < min_wh) | (x[:, 2:4] > max_wh)).any(1), 4] = 0  # width-height
        x = x.transpose(0, -1)[xc[xi]]  # confidence

        # Cat apriori labels if autolabelling
        if labels and len(labels[xi]):
            lb = labels[xi]
            v = torch.zeros((len(lb), nc + nm + 5), device=x.device)
            v[:, :4] = lb[:, 1:5]  # box
            v[range(len(lb)), lb[:, 0].long() + 4] = 1.0  # cls
            x = torch.cat((x, v), 0)

        # If none remain process next image
        if not x.shape[0]:
            continue

        # Detections matrix nx6 (xyxy, conf, cls)
        box, cls, mask = x.split((4, nc, nm), 1)
        box = xywh2xyxy(box)  # center_x, center_y, width, height) to (x1, y1, x2, y2)
        if multi_label:
            i, j = (cls > conf_thres).nonzero(as_tuple=False).T
            x = torch.cat((box[i], x[i, 4 + j, None], j[:, None].float(), mask[i]), 1)
        else:  # best class only
            conf, j = cls.max(1, keepdim=True)
            x = torch.cat((box, conf, j.float(), mask), 1)[conf.view(-1) > conf_thres]

        # Filter by class
        if classes is not None:
            x = x[(x[:, 5:6] == torch.tensor(classes, device=x.device)).any(1)]

        # Apply finite constraint
        # if not torch.isfinite(x).all():
        #     x = x[torch.isfinite(x).all(1)]

        # Check shape
        n = x.shape[0]  # number of boxes
        if not n:  # no boxes
            continue
        x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence and remove excess boxes

        # Batched NMS
        c = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
        boxes, scores = x[:, :4] + c, x[:, 4]  # boxes (offset by class), scores
        i = torchvision.ops.nms(boxes, scores, iou_thres)  # NMS
        i = i[:max_det]  # limit detections
        if merge and (1 < n < 3E3):  # Merge NMS (boxes merged using weighted mean)
            # Update boxes as boxes(i,4) = weights(i,n) * boxes(n,4)
            iou = box_iou(boxes[i], boxes) > iou_thres  # iou matrix
            weights = iou * scores[None]  # box weights
            x[i, :4] = torch.mm(weights, x[:, :4]).float() / weights.sum(1, keepdim=True)  # merged boxes
            if redundant:
                i = i[iou.sum(1) > 1]  # require redundancy

        output[xi] = x[i]
        if mps:
            output[xi] = output[xi].to(device)
        if (time.time() - t) > time_limit:
            logger.warning(f'WARNING ⚠️ NMS time limit {time_limit:.3f}s exceeded')
            break  # time limit exceeded

    return output

def xywh2xyxy(x):
    """
    Convert bounding box coordinates from (x, y, width, height) format to (x1, y1, x2, y2) format where (x1, y1) is the
    top-left corner and (x2, y2) is the bottom-right corner.

    Args:
        x (np.ndarray | torch.Tensor): The input bounding box coordinates in (x, y, width, height) format.
    Returns:
        y (np.ndarray | torch.Tensor): The bounding box coordinates in (x1, y1, x2, y2) format.
    """
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = x[..., 0] - x[..., 2] / 2  # top left x
    y[..., 1] = x[..., 1] - x[..., 3] / 2  # top left y
    y[..., 2] = x[..., 0] + x[..., 2] / 2  # bottom right x
    y[..., 3] = x[..., 1] + x[..., 3] / 2  # bottom right y
    return y

def xyxy2xywh(x):
    """
    Convert bounding box coordinates from (x1, y1, x2, y2) format to (x, y, width, height) format.

    Args:
        x (np.ndarray | torch.Tensor): The input bounding box coordinates in (x1, y1, x2, y2) format.
    Returns:
       y (np.ndarray | torch.Tensor): The bounding box coordinates in (x, y, width, height) format.
    """
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = (x[..., 0] + x[..., 2]) / 2  # x center
    y[..., 1] = (x[..., 1] + x[..., 3]) / 2  # y center
    y[..., 2] = x[..., 2] - x[..., 0]  # width
    y[..., 3] = x[..., 3] - x[..., 1]  # height
    return y

def box_iou(box1, box2, eps=1e-7):
    """
    Calculate intersection-over-union (IoU) of boxes.
    Both sets of boxes are expected to be in (x1, y1, x2, y2) format.
    Based on https://github.com/pytorch/vision/blob/master/torchvision/ops/boxes.py

    Args:
        box1 (torch.Tensor): A tensor of shape (N, 4) representing N bounding boxes.
        box2 (torch.Tensor): A tensor of shape (M, 4) representing M bounding boxes.
        eps (float, optional): A small value to avoid division by zero. Defaults to 1e-7.

    Returns:
        (torch.Tensor): An NxM tensor containing the pairwise IoU values for every element in box1 and box2.
    """

    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    (a1, a2), (b1, b2) = box1.unsqueeze(1).chunk(2, 2), box2.unsqueeze(0).chunk(2, 2)
    inter = (torch.min(a2, b2) - torch.max(a1, b1)).clamp_(0).prod(2)

    # IoU = inter / (area1 + area2 - inter)
    return inter / ((a2 - a1).prod(2) + (b2 - b1).prod(2) - inter + eps)

def process_mask_native(protos, masks_in, bboxes, shape):
    """
    It takes the output of the mask head, and crops it after upsampling to the bounding boxes.

    Args:
      protos (torch.Tensor): [mask_dim, mask_h, mask_w]
      masks_in (torch.Tensor): [n, mask_dim], n is number of masks after nms
      bboxes (torch.Tensor): [n, 4], n is number of masks after nms
      shape (tuple): the size of the input image (h,w)

    Returns:
      masks (torch.Tensor): The returned masks with dimensions [h, w, n]
    """
    c, mh, mw = protos.shape  # CHW
    masks = (masks_in @ protos.float().view(c, -1)).sigmoid().view(-1, mh, mw)
    gain = min(mh / shape[0], mw / shape[1])  # gain  = old / new
    pad = (mw - shape[1] * gain) / 2, (mh - shape[0] * gain) / 2  # wh padding
    top, left = int(pad[1]), int(pad[0])  # y, x
    bottom, right = int(mh - pad[1]), int(mw - pad[0])
    masks = masks[:, top:bottom, left:right]

    masks = F.interpolate(masks[None], shape, mode='bilinear', align_corners=False)[0]  # CHW
    masks = crop_mask(masks, bboxes)  # CHW
    return masks.gt_(0.5)

def crop_mask(masks, boxes):
    """
    It takes a mask and a bounding box, and returns a mask that is cropped to the bounding box

    Args:
      masks (torch.Tensor): [h, w, n] tensor of masks
      boxes (torch.Tensor): [n, 4] tensor of bbox coordinates in relative point form

    Returns:
      (torch.Tensor): The masks are being cropped to the bounding box.
    """
    n, h, w = masks.shape
    x1, y1, x2, y2 = torch.chunk(boxes[:, :, None], 4, 1)  # x1 shape(n,1,1)
    r = torch.arange(w, device=masks.device, dtype=x1.dtype)[None, None, :]  # rows shape(1,1,w)
    c = torch.arange(h, device=masks.device, dtype=x1.dtype)[None, :, None]  # cols shape(1,h,1)

    return masks * ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))



"""ExternalRepresentation module"""
from __future__ import annotations
from typing import Callable
import numpy as np
from overrides import overrides

from vre.utils import image_resize_batch
from .representation import Representation, ReprOut

class ExternalRepresentation(Representation):
    """
    A placeholder for external data that we cannot compute (i.e., sfm, edges_gb etc.). Must provide a plot fn and
    optionally a resize_fn.
    """
    def __init__(self, name: str, make_image_fn: Callable,
                 resize_fn: Callable[[ReprOut, tuple[int, int]], ReprOut] | None = None):
        super().__init__(name=name, dependencies=[])
        assert isinstance(make_image_fn, Callable), type(make_image_fn)
        assert isinstance(resize_fn, (Callable, type(None))), type(resize_fn)
        self.make_image_fn = make_image_fn
        self.resize_fn = resize_fn

    @overrides
    def make(self, frames: np.ndarray, dep_data: dict[str, ReprOut] | None = None) -> ReprOut:
        raise ValueError(f"[{self}] ExternalRepresentation.make() cannot be called. Data must be loaded externally")

    @overrides
    def make_images(self, frames: np.ndarray, repr_data: ReprOut) -> np.ndarray:
        return self.make_image_fn(frames, repr_data)

    @overrides
    def size(self, repr_data: ReprOut) -> tuple[int, int]:
        return repr_data.output.shape[1:3]

    @overrides
    def resize(self, repr_data: ReprOut, new_size: tuple[int, int]) -> ReprOut:
        if self.resize_fn is None:
            return ReprOut(output=image_resize_batch(repr_data.output, height=new_size[0], width=new_size[1]))
        return self.resize_fn(repr_data, new_size)



# pylint: disable=all
import torch as tr
import torch.nn as nn
import torch.nn.functional as F

def conv(dIn, dOut, kernel_size, padding, stride, dilation):
    return nn.Sequential(
        nn.Conv2d(in_channels=dIn, out_channels=dOut, kernel_size=kernel_size, padding=padding, \
            stride=stride, dilation=dilation),
        nn.BatchNorm2d(num_features=dOut),
        nn.ReLU(inplace=True)
    )

def conv_tr(dIn, dOut, kernel_size, padding, stride, dilation):
    return nn.Sequential(
        nn.ConvTranspose2d(in_channels=dIn, out_channels=dOut, kernel_size=kernel_size, padding=padding, \
            stride=stride, dilation=dilation),
        nn.BatchNorm2d(num_features=dOut),
        nn.ReLU(inplace=True)
    )

def fCat(a, b):
    diffUp, diffLeft = b.shape[-2] - a.shape[-2], b.shape[-1] - a.shape[-1]
    a = F.pad(a, (0, diffLeft, 0, diffUp))
    c = tr.cat([a, b], dim=1)
    return c

class EncoderMap2Map(nn.Module):
    def __init__(self, dIn, numFilters=16):
        self.dIn = dIn
        self.numFilters = numFilters
        super().__init__()

        self.conv1 = conv(dIn=dIn, dOut=numFilters, kernel_size=3, padding=1, stride=1, dilation=1)
        self.conv2 = conv(dIn=numFilters, dOut=numFilters, kernel_size=3, padding=1, stride=1, dilation=1)
        self.conv3 = conv(dIn=numFilters, dOut=numFilters, kernel_size=3, padding=1, stride=2, dilation=1)

        self.conv4 = conv(dIn=numFilters, dOut=numFilters * 2, kernel_size=3, padding=1, stride=1, dilation=1)
        self.conv5 = conv(dIn=numFilters * 2, dOut=numFilters * 2, kernel_size=3, padding=1, stride=1, dilation=1)
        self.conv6 = conv(dIn=numFilters * 2, dOut=numFilters * 2, kernel_size=3, padding=1, stride=2, dilation=1)

        self.conv7 = conv(dIn=numFilters * 2, dOut=numFilters * 4, kernel_size=3, padding=1, stride=1, dilation=1)
        self.conv8 = conv(dIn=numFilters * 4, dOut=numFilters * 4, kernel_size=3, padding=1, stride=1, dilation=1)
        self.conv9 = conv(dIn=numFilters * 4, dOut=numFilters * 4, kernel_size=3, padding=1, stride=2, dilation=1)

        self.dilate1 = conv(dIn=numFilters * 4, dOut=numFilters * 8, kernel_size=3, padding=1, stride=1, dilation=1)
        self.dilate2 = conv(dIn=numFilters * 8, dOut=numFilters * 8, kernel_size=3, padding=2, stride=1, dilation=2)
        self.dilate3 = conv(dIn=numFilters * 8, dOut=numFilters * 8, kernel_size=3, padding=4, stride=1, dilation=4)
        self.dilate4 = conv(dIn=numFilters * 8, dOut=numFilters * 8, kernel_size=3, padding=8, stride=1, dilation=8)
        self.dilate5 = conv(dIn=numFilters * 8, dOut=numFilters * 8, kernel_size=3, padding=16, stride=1, dilation=16)
        self.dilate6 = conv(dIn=numFilters * 8, dOut=numFilters * 8, kernel_size=3, padding=32, stride=1, dilation=32)

        self.conv_transpose10 = conv_tr(dIn=numFilters * 8, dOut=numFilters * 4, kernel_size=3, \
            padding=1, stride=2, dilation=1)
        self.conv11 = conv(dIn=numFilters * 4 * 2, dOut=numFilters * 4, kernel_size=3, padding=1, stride=1, dilation=1)
        self.conv_transpose12 = conv_tr(dIn=numFilters * 4, dOut=numFilters * 2, kernel_size=3, \
            padding=1, stride=2, dilation=1)
        self.conv13 = conv(dIn=numFilters * 2 * 2, dOut=numFilters, kernel_size=3, padding=1, stride=1, dilation=1)
        self.conv_transpose14 = conv_tr(dIn=numFilters, dOut=numFilters, kernel_size=3, \
            padding=1, stride=2, dilation=1)
        self.conv15 = conv(dIn=numFilters * 2, dOut=numFilters, kernel_size=3, padding=1, stride=1, dilation=1)

    def forward(self, x: tr.Tensor) -> tr.Tensor:
        # x::dInxHxW
        y_1 = self.conv1(x)
        y_2 = self.conv2(y_1)
        y_3 = self.conv3(y_2)
        y_4 = self.conv4(y_3)
        y_5 = self.conv5(y_4)
        y_6 = self.conv6(y_5)

        y_7 = self.conv7(y_6)
        y_8 = self.conv8(y_7)
        y_9 = self.conv9(y_8)

        y_dilate1 = self.dilate1(y_9)
        y_dilate2 = self.dilate2(y_dilate1)
        y_dilate3 = self.dilate3(y_dilate2)
        y_dilate4 = self.dilate4(y_dilate3)
        y_dilate5 = self.dilate5(y_dilate4)
        y_dilate6 = self.dilate6(y_dilate5)
        y_dilate_sum = y_dilate1 + y_dilate2 + y_dilate3 + y_dilate4 + y_dilate5 + y_dilate6

        y_10 = fCat(self.conv_transpose10(y_dilate_sum), y_7)
        y_11 = self.conv11(y_10)
        y_12 = fCat(self.conv_transpose12(y_11), y_4)
        y_13 = self.conv13(y_12)
        y_14 = fCat(self.conv_transpose14(y_13), y_1)
        y_15 = self.conv15(y_14)
        return y_15

    def __str__(self):
        return f"Encoder Map2Map. dIn: {self.dIn}. NF: {self.numFilters}."

class DecoderMap2Map(nn.Module):
    def __init__(self, dOut, numFilters=16):
        self.dOut = dOut
        self.numFilters = numFilters
        super().__init__()

        self.decoderConv = nn.Conv2d(in_channels=numFilters, out_channels=dOut, kernel_size=1, \
            padding=0, stride=1, dilation=1)

    def forward(self, x):
        y = self.decoderConv(x)
        return y

    def __str__(self):
        return f"Decoder Map2Map. dOut: {self.dOut}. NF: {self.numFilters}."



# pylint: disable=all
import torch
import torch.nn as nn
import torch.nn.functional as F

from .update import BasicUpdateBlock, SmallUpdateBlock
from .extractor import BasicEncoder, SmallEncoder
from .corr import CorrBlock, AlternateCorrBlock
from .utils import coords_grid, upflow8
from torch.amp import autocast


class RAFT(nn.Module):
    def __init__(self, args):
        super(RAFT, self).__init__()
        self.args = args

        if args.small:
            self.hidden_dim = hdim = 96
            self.context_dim = cdim = 64
            args.corr_levels = 4
            args.corr_radius = 3
        else:
            self.hidden_dim = hdim = 128
            self.context_dim = cdim = 128
            args.corr_levels = 4
            args.corr_radius = 4

        self.args.dropout = 0

        self.args.alternate_corr = False

        # feature network, context network, and update block
        if args.small:
            self.fnet = SmallEncoder(output_dim=128, norm_fn='instance', dropout=args.dropout)
            self.cnet = SmallEncoder(output_dim=hdim+cdim, norm_fn='none', dropout=args.dropout)
            self.update_block = SmallUpdateBlock(self.args, hidden_dim=hdim)

        else:
            self.fnet = BasicEncoder(output_dim=256, norm_fn='instance', dropout=args.dropout)
            self.cnet = BasicEncoder(output_dim=hdim+cdim, norm_fn='batch', dropout=args.dropout)
            self.update_block = BasicUpdateBlock(self.args, hidden_dim=hdim)

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def initialize_flow(self, img):
        """ Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H//8, W//8).to(img.device)
        coords1 = coords_grid(N, H//8, W//8).to(img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """ Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination """
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3,3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8*H, 8*W)


    def forward(self, image1, image2, iters=12, flow_init=None, upsample=True, test_mode=False):
        """ Estimate optical flow between pair of frames """

        image1 = 2 * (image1 / 255.0) - 1.0
        image2 = 2 * (image2 / 255.0) - 1.0

        image1 = image1.contiguous()
        image2 = image2.contiguous()

        hdim = self.hidden_dim
        cdim = self.context_dim

        # run the feature network
        with autocast("cuda", enabled=self.args.mixed_precision):
            fmap1, fmap2 = self.fnet([image1, image2])

        fmap1 = fmap1.float()
        fmap2 = fmap2.float()
        if self.args.alternate_corr:
            corr_fn = AlternateCorrBlock(fmap1, fmap2, radius=self.args.corr_radius)
        else:
            corr_fn = CorrBlock(fmap1, fmap2, radius=self.args.corr_radius)

        # run the context network
        with autocast("cuda", enabled=self.args.mixed_precision):
            cnet = self.cnet(image1)
            net, inp = torch.split(cnet, [hdim, cdim], dim=1)
            net = torch.tanh(net)
            inp = torch.relu(inp)

        coords0, coords1 = self.initialize_flow(image1)

        if flow_init is not None:
            coords1 = coords1 + flow_init

        flow_predictions = []
        for itr in range(iters):
            coords1 = coords1.detach()
            corr = corr_fn(coords1) # index correlation volume

            flow = coords1 - coords0
            with autocast("cuda", enabled=self.args.mixed_precision):
                net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = upflow8(coords1 - coords0)
            else:
                flow_up = self.upsample_flow(coords1 - coords0, up_mask)

            flow_predictions.append(flow_up)

        if test_mode:
            return coords1 - coords0, flow_up
        return flow_predictions



"""init file"""
from .stored_represntation import StoredRepresentation, NpzRepresentation
from .normed_representation import NormedRepresentation
from .task_mapper import TaskMapper
from .cv_representations import (RGBRepresentation, HSVRepresentation, EdgesRepresentation, SemanticRepresentation,
                                 DepthRepresentation, OpticalFlowRepresentation, NormalsRepresentation,
                                 ColorRepresentation)



"""setup.py"""
from pathlib import Path
from setuptools import setup, find_packages

NAME = "video-representations-extractor"
VERSION = "1.1.6"
DESCRIPTION = "Video Representations Extractor (VRE) for computing algorithmic or neural representations of each frame."
URL = "https://gitlab.com/video-representations-extractor/video-representations-extractor"

with open(f"{Path(__file__).absolute().parent}/README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

REQUIRED = [
    "numpy>=1.21.6,<2.0.0",
    "PyYAML==6.0",
    "pims==0.7.0",
    "tqdm==4.66.5",
    "natsort==8.2.0",
    "torch==2.4.1",
    "torchvision==0.19.1",
    "overrides==7.7.0",
    "pandas>=2.1.3,<3.0.0",
    "matplotlib==3.7.1",
    "loggez==0.4.1",
    "omegaconf==2.3.0",
    "lovely_tensors==0.1.17",
    "opencv-python==4.7.0.68",
    "moviepy==1.0.3",
    "Pillow==10.3.0",
    "pycocotools==2.0.7",
    "flow_vis==0.1",
    "timm==1.0.9",
    "fvcore==0.1.5.post20221221",
    "diffusers==0.30.3",
]

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    packages=find_packages(),
    install_requires=REQUIRED,
    dependency_links=[],
    license="MIT",
    python_requires=">=3.10",
    scripts=["bin/vre", "bin/vre_collage"],
)



"""FlowRaft representation"""
from pathlib import Path
from overrides import overrides
import numpy as np
import torch as tr
from torch.nn import functional as F
from vre.representations import Representation, ReprOut, LearnedRepresentationMixin
from vre.utils import image_resize_batch, fetch_weights, vre_load_weights, VREVideo, colorize_optical_flow
from vre.logger import vre_logger as logger

try:
    from .raft_impl import RAFT, InputPadder
except ImportError:
    from raft_impl import RAFT, InputPadder

class FlowRaft(Representation, LearnedRepresentationMixin):
    """FlowRaft representation"""
    def __init__(self, inference_height: int, inference_width: int, iters: int, small: bool,
                 seed: int | None = None, flow_delta_frames: int = 1, **kwargs):
        super().__init__(**kwargs)
        assert inference_height >= 128 and inference_width >= 128, f"This flow doesn't work with small " \
            f"videos. At least 128x128 is required, but got {inference_height}x{inference_width}"
        self.mixed_precision = False
        self.small = small
        self.iters = iters
        self.seed = seed
        self.flow_delta_frames = flow_delta_frames

        self.model: RAFT | None = None
        self.inference_width = inference_width
        self.inference_height = inference_height

    @overrides
    def vre_setup(self, load_weights: bool = True):
        tr.manual_seed(self.seed) if self.seed is not None else None
        self.model = RAFT(self).to("cpu")
        if load_weights:
            def convert(data: dict[str, tr.Tensor]) -> dict[str, tr.Tensor]:
                logger.warning("REMOVE THIS WHEN THERE'S TIME")
                return {k.replace("module.", ""): v for k, v in data.items()}
            raft_things_path = fetch_weights(__file__) / "raft-things.ckpt"
            self.model.load_state_dict(convert(vre_load_weights(raft_things_path)))
        self.model = self.model.eval().to(self.device)

    @overrides
    def vre_dep_data(self, video: VREVideo, ixs: slice | list[int], output_dir: Path | None) -> dict[str, ReprOut]:
        ixs: list[int] = list(range(ixs.start, ixs.stop)) if isinstance(ixs, slice) else ixs
        right_ixs = [min(ix + self.flow_delta_frames, len(video)) for ix in ixs]
        right_frames = np.array(video[right_ixs]) # godbless pims I hope it caches this properly
        return {"right_frames": ReprOut(output=right_frames)}

    @overrides
    def make(self, frames: np.ndarray, dep_data: dict[str, ReprOut] | None = None) -> ReprOut:
        right_frames = dep_data["right_frames"]
        self._check_frames_resolution_requirements(frames)
        self._check_frames_resolution_requirements(right_frames)

        source, dest = self._preprocess(frames), self._preprocess(right_frames)
        tr.manual_seed(self.seed) if self.seed is not None else None
        with tr.no_grad():
            _, predictions = self.model(source, dest, iters=self.iters, test_mode=True)
        flow = self._postporcess(predictions)
        return ReprOut(output=flow)

    @overrides
    def make_images(self, frames: np.ndarray, repr_data: ReprOut) -> np.ndarray:
        return np.array([colorize_optical_flow(_pred) for _pred in repr_data.output])

    @overrides
    def size(self, repr_data: ReprOut) -> tuple[int, int]:
        return repr_data.output.shape[1:3]

    @overrides
    def resize(self, repr_data: ReprOut, new_size: tuple[int, int]) -> ReprOut:
        return ReprOut(output=image_resize_batch(repr_data.output, *new_size))

    def _check_frames_resolution_requirements(self, frames: np.ndarray):
        assert frames.shape[1] >= 128 and frames.shape[2] >= 128, \
            f"This flow doesn't work with small videos. At least 128x128 is required, but got {frames.shape}"
        assert frames.shape[1] >= self.inference_height and frames.shape[2] >= self.inference_width, \
            f"{frames.shape} vs {self.inference_height}x{self.inference_width}"

    def _preprocess(self, frames: np.ndarray) -> tuple[tr.Tensor, tr.Tensor]:
        tr_frames = tr.from_numpy(frames).to(self.device)
        tr_frames = tr_frames.permute(0, 3, 1, 2).float() # (B, C, H, W)
        frames_rsz = F.interpolate(tr_frames, (self.inference_height, self.inference_width), mode="bilinear")
        padder = InputPadder((len(frames), 3, self.inference_height, self.inference_width))
        frames_padded = padder.pad(frames_rsz)
        return frames_padded

    def _postporcess(self, predictions: tr.Tensor) -> np.ndarray:
        padder = InputPadder((len(predictions), 3, self.inference_height, self.inference_width))
        flow_unpad = padder.unpad(predictions).cpu().numpy()
        flow_perm = flow_unpad.transpose(0, 2, 3, 1)
        flow_unpad_norm = flow_perm / (self.inference_height, self.inference_width) # [-1 : 1]
        flow_unpad_norm = flow_unpad_norm.astype(np.float16)
        return flow_unpad_norm

    def vre_free(self):
        if str(self.device).startswith("cuda"):
            self.model.to("cpu")
            tr.cuda.empty_cache()



"""Dexined representation."""
import numpy as np
import torch as tr
from overrides import overrides
from vre.representations import Representation, ReprOut, LearnedRepresentationMixin
from vre.logger import vre_logger as logger
from vre.utils import image_resize_batch, fetch_weights, vre_load_weights

try:
    from .model_dexined import DexiNed as Model
except ImportError:
    from model_dexined import DexiNed as Model

class DexiNed(Representation, LearnedRepresentationMixin):
    """Dexined representation."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.model: Model | None = None
        self.inference_height, self.inference_width = 512, 512 # fixed for this model

    @overrides
    def vre_setup(self, load_weights: bool = True):
        self.model = Model().eval()
        if load_weights:
            self.model.load_state_dict(vre_load_weights(fetch_weights(__file__) / "dexined.pth"))
        self.model = self.model.to(self.device)

    @overrides
    def make(self, frames: np.ndarray, dep_data: dict[str, ReprOut] | None = None) -> ReprOut:
        tr_frames = self._preprocess(frames, self.inference_height, self.inference_width)
        with tr.no_grad():
            y = self.model.forward(tr_frames)
        outs = self._postprocess(y)
        return ReprOut(output=outs)

    @overrides
    def make_images(self, frames: np.ndarray, repr_data: ReprOut) -> np.ndarray:
        x = np.repeat(np.expand_dims(repr_data.output, axis=-1), 3, axis=-1)
        return (x * 255).astype(np.uint8)

    @overrides
    def size(self, repr_data: ReprOut) -> tuple[int, int]:
        return repr_data.output.shape[1:3]

    @overrides
    def resize(self, repr_data: ReprOut, new_size: tuple[int, int]) -> ReprOut:
        return ReprOut(output=image_resize_batch(repr_data.output, *new_size))

    @overrides
    def vre_free(self):
        if str(self.device).startswith("cuda"):
            self.model.to("cpu")
            tr.cuda.empty_cache()

    def _preprocess(self, images: np.ndarray, height: int, width: int) -> tr.Tensor:
        assert len(images.shape) == 4, images.shape
        logger.debug2(f"Original shape: {images.shape}")
        images_resize = image_resize_batch(images, height, width)
        mean_pixel_values = [103.939, 116.779, 123.68]
        images_norm = images_resize.astype(np.float32) - mean_pixel_values
        # N, H, W, C -> N, C, H, W
        images_norm = images_norm.transpose((0, 3, 1, 2))
        tr_images_norm = tr.from_numpy(images_norm).to(self.device).float()
        return tr_images_norm

    def _postprocess(self, y: list[tr.Tensor]) -> np.ndarray:
        # y :: (B, T, H, W)
        y_cat = tr.cat(y, dim=1)
        assert len(y_cat.shape) == 4, y_cat.shape
        y_s = y_cat.sigmoid()
        A = y_s.min(dim=-1, keepdims=True)[0].min(dim=-2, keepdims=True)[0]
        B = y_s.max(dim=-1, keepdims=True)[0].max(dim=-2, keepdims=True)[0]
        y_s_normed: tr.Tensor = 1 - (y_s - A) * (B - A)
        y_s_final = y_s_normed.mean(dim=1).cpu().numpy().astype(np.float32)
        return y_s_final



"""common visualizing functions for standard representations, like depth, semantic, optical flow etc."""
import matplotlib
import flow_vis
import numpy as np
from .colorized_sem_seg import ColorizerSemSeg

def colorize_depth(depth_map: np.ndarray, min_depth: float, max_depth: float, cmap="Spectral") -> np.ndarray:
    """Colorize depth maps. Returns a [0:1] (H, W, 3) float image."""
    assert len(depth_map.shape) == 3 or (len(depth_map.shape) == 4 and depth_map.shape[-1] == 1), depth_map.shape
    assert isinstance(depth_map, np.ndarray), depth_map
    cm = matplotlib.colormaps[cmap]
    depth = ((depth_map - min_depth) / (max_depth - min_depth)).clip(0, 1)
    img_colored_np = cm(depth, bytes=False)[..., 0:3]  # value from 0 to 1
    return img_colored_np.astype(np.float32)

def colorize_optical_flow(flow_map: np.ndarray) -> np.ndarray:
    """Colorize optical flow maps. returns a [0:255] (H, W, 3) uint8 image."""
    assert len(flow_map.shape) == 3 and flow_map.shape[2] == 2, flow_map.shape
    return flow_vis.flow_to_color(flow_map)

def colorize_semantic_segmentation(semantic_map: np.ndarray, classes: list[str], color_map: list[tuple[int, int, int]],
                                   original_rgb: np.ndarray | None, font_size_scale: float=1, alpha: float = 0.8):
    """Colorize asemantic segmentation maps. Must be argmaxed (H, W). Can paint over the original RGB frame or not."""
    assert np.issubdtype(semantic_map.dtype, np.integer), semantic_map.dtype
    assert (max_class := semantic_map.max()) <= len(color_map), (max_class, len(color_map))
    assert len(semantic_map.shape) == 2, semantic_map.shape
    rgb = original_rgb if original_rgb is not None else np.zeros((*semantic_map.shape, 3), dtype=np.uint8)
    alpha = alpha if original_rgb is not None else 1
    return ColorizerSemSeg(rgb, classes, color_map, font_size_scale, alpha).draw_sem_seg(semantic_map).get_image()



"""Helper module to make sense of the arguments sent to vre.run()"""
from typing import Any
from pathlib import Path
import shutil

from .utils import is_dir_empty, parsed_str_type, VREVideo
from .logger import vre_logger as logger

RepresentationsSetup = dict[str, dict[str, Any]]

class VRERuntimeArgs:
    """
    VRE runtime args. Helper class to process the arguments sent to vre.run()
    Parameters:
    - video The video that this run operates on
    - representations The dictionary of representations that this run operates on
    - output_dir The directory used to output representations in this run.
    - start_frame The first frame to process (inclusive). If not provided, defaults to 0.
    - end_frame The last frame to process (inclusive). If not provided, defaults to len(video).
    - batch_size The batch size to use when processing the video. If not provided, defaults to 1.
    - export_npy Whether to export the npy files
    - export_png Whether to export the png files
    - output_dir_exists_mode What to do if the output dir already exists. Can be one of:
        - 'overwrite' Overwrite the output dir if it already exists
        - 'skip_computed' Skip the computed frames and continue from the last computed frame
        - 'raise' (default) Raise an error if the output dir already exists
    - exception_mode What to do when encountering an exception. It always writes the exception to 'exception.txt'.
        - 'skip_representation' Will stop the run of the current representation and start the next one
        - 'stop_execution' (default) Will stop the execution of VRE
    - output_size The resulted output shape in the npy/png directories. Valid options: a tuple (h, w), or a string:
        - 'native' whatever each representation outputs out of the box)
        - 'video_shape' (default) resizing to the video shape
    - n_threads_data_storer The number of workers used for the ThreadPool that stores data at each step. This is
    needed because storing data takes a lot of time sometimes, even more than the computation itself. Default: 1.
    - load_from_disk_if_computed If true, then it will try to read from the disk if a representation is computed.
    """
    def __init__(self, video: VREVideo, representations: dict[str, "Representation"], output_dir: Path,
                 start_frame: int | None, end_frame: int | None, batch_size: int, export_npy: bool, export_png: bool,
                 output_dir_exists_mode: str, exception_mode: str, output_size: str | tuple,
                 n_threads_data_storer: int, load_from_disk_if_computed: bool):
        assert batch_size >= 1, f"batch size must be >= 1, got {batch_size}"
        assert export_npy + export_png > 0, "At least one of export modes must be True"
        assert output_dir_exists_mode in ("overwrite", "skip_computed", "raise"), output_dir_exists_mode
        assert exception_mode in ("stop_execution", "skip_representation"), exception_mode
        if isinstance(output_size, str):
            assert output_size in ("native", "video_shape"), output_size
        else:
            assert len(output_size) == 2 and all(isinstance(x, int) for x in output_size), output_size
        if start_frame is None:
            start_frame = 0
            logger.warning("start frame not set, default to 0")
        if end_frame is None:
            logger.warning(f"end frame not set, default to the last frame of the video: {len(video)}")
            end_frame = len(video)

        assert isinstance(start_frame, int) and start_frame <= end_frame, (start_frame, end_frame)
        self.video = video
        self.output_dir = output_dir
        self.start_frame = start_frame
        self.end_frame = end_frame
        self.batch_size = batch_size
        self.export_npy = export_npy
        self.export_png = export_png
        self.output_dir_exists_mode = output_dir_exists_mode
        self.exception_mode = exception_mode
        self.output_size = tuple(output_size) if not isinstance(output_size, str) else output_size
        self.n_threads_data_storer = n_threads_data_storer
        self.representations = representations
        self.load_from_disk_if_computed = load_from_disk_if_computed

        self.batch_sizes = {k: batch_size if r.batch_size is None else r.batch_size
                            for k, r in representations.items()}
        self.output_sizes = {k: output_size if r.output_size is None else r.output_size
                             for k, r in representations.items()}

        self.npy_paths: dict[str, list[Path]] = self._make_npy_paths() # {repr: [out_dir/npy/0.npz, ...]}
        self.png_paths: dict[str, list[Path]] = self._make_png_paths() # {repr: [out_dir/png/0.png, ...]}
        assert len(self.npy_paths) > 0 and len(self.png_paths) > 0, (len(self.npy_paths), len(self.png_paths))
        self._make_and_check_dirs()
        logger.info(self)

    def _make_npy_paths(self) -> dict[str, list[Path]]:
        npy_paths = {}
        for name in self.representations.keys():
            npy_base_dir = self.output_dir / name / "npy/"
            npy_paths[name] = [npy_base_dir / f"{t}.npz" for t in range(len(self.video))]
        return npy_paths

    def _make_png_paths(self) -> dict[str, list[Path]]:
        png_paths = {}
        for name in self.representations.keys():
            png_base_dir = self.output_dir / name / "png/"
            png_paths[name] = [png_base_dir / f"{t}.png" for t in range(len(self.video))]
        return png_paths

    def _make_and_check_dirs(self):
        for representation in self.representations:
            npy_dir: Path = self.npy_paths[representation][0].parent
            png_dir: Path = self.png_paths[representation][0].parent

            if self.export_npy and npy_dir.exists() and not is_dir_empty(npy_dir, "*.npz"):
                if self.output_dir_exists_mode == "overwrite":
                    logger.debug(f"Output dir '{npy_dir}' already exists, will overwrite it")
                    shutil.rmtree(npy_dir)
                else:
                    assert self.output_dir_exists_mode != "raise", \
                        f"'{npy_dir}' exists. Set --output_dir_exists_mode to 'overwrite' or 'skip_computed'"
            npy_dir.mkdir(exist_ok=True, parents=True)

            if self.export_png and png_dir.exists() and not is_dir_empty(png_dir, "*.png"):
                if self.output_dir_exists_mode == "overwrite":
                    logger.debug(f"Output dir '{png_dir}' already exists, will overwrite it")
                    shutil.rmtree(png_dir)
                else:
                    assert self.output_dir_exists_mode != "raise", \
                        f"'{png_dir}' exists. Set --output_dir_exists_mode to 'overwrite' or 'skip_computed'"
            png_dir.mkdir(exist_ok=True, parents=True)

    def __repr__(self):
        return f"""[{parsed_str_type(self)}]
- Video path: '{getattr(self.video, "file", "")}'
- Output dir: '{self.output_dir}')
- Representations ({len(self.representations)}): {", ".join(x for x in self.representations.keys())}
- Video shape: {self.video.shape} (FPS: {self.video.frame_rate:.2f})
- Output frames ({self.end_frame - self.start_frame}): [{self.start_frame} : {self.end_frame - 1}]
- Output shape: {self.output_size if self.output_size != "video_shape" else self.video.frame_shape[0:2]}
- Batch size: {self.batch_size}
- Export npy: {self.export_npy}
- Export png: {self.export_png}
- Exception mode: '{self.exception_mode}'
- Output dir exists mode: '{self.output_dir_exists_mode}'
- Thread pool workers for storing data (0 = using main thread): {self.n_threads_data_storer}
- Load from disk if computed: {self.load_from_disk_if_computed}
"""



"""
Helper mixin class that adds the weights/device relevant methods & properties for such representation
A representation that inherits this also must have weights in the weights repository.
"""
from abc import abstractmethod, ABC
import torch as tr

VREDevice = str | tr.device # not only torch, but this is what we support atm

class LearnedRepresentationMixin(ABC):
    """Learned Representastion Mixin for VRE implementation"""
    def __init__(self):
        self.device: VREDevice = "cpu"

    # TODO: make this and get rid of the hardcoded stuff in utils.
    # @property
    # @abstractmethod
    # def weights_files(self) -> list[str]:
    #     """
    #     A list of files that must be present on the disk or downloaded from the weights repository. Only the stem is
    #     needed and vre_setup() must handle their loading. The data is stored under {weights_dir}/{repr_type}/[names]
    #     For example: 'depth/dpt' has 'depth_dpt_midas.pth' and is stored at 'weights_dir/depth/dpt/depth_dpt_midas.pth
    #     """

    @abstractmethod
    def vre_setup(self, load_weights: bool = True):
        """
        Setup method for this representation. This is required to run this representation from within VRE. We do this
        setup separately, so we can instatiate the object without doing any VRE specific setup, like loading weights.
        """

    @abstractmethod
    def vre_free(self):
        """Needed to deallocate stuff from cuda mostly. After this, you need to run vre_setup() again."""
